;**********************************************************************
;      PO Metal detector using PIC12F1840                             *
;                                                                     *
;    Filename:                  metal1.asm                            *
;    Date:                      10.01.2013                            *
;    File Version:              1.25                                  *
;                                                                     *
;   Author:Teemo Berkis                                               *
;       www.digiwood.ee                                               *
;                                                                     *
;    Written and assembled under MPLAB Version 8.83                   *
;                                                                     *
;    Before assembling code set correct PIC processor in MPLAB        *
;    from menu option Configure - Select Device                       *
;                                                                     *
;                                                                     *
;**********************************************************************
; Pinout connections 
;                            ___ ___
;                      Vdd+ -|1 ^ 8|- Vss-
;             mosfet drive  -|2   7|- Piezo1/RS232_gnd
;           measuring input -|3   6|- LED output
;         pushbutton switch -|4   5|- Piezo2/RS232_TX
;                            -------
;

	list	p=12F1840
	radix	dec
	include	"p12f1840.inc"	
	include	"Metalmacros.inc"
	
		errorlevel	-302	; Dont complain about BANK 1 Registers during assembly

;==========================================================================
;
;       Configuration Bits
;
;==========================================================================
 ;Internal osc.

    __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOREN_NSLEEP & _CLKOUTEN_OFF & _IESO_OFF & _FCMEN_OFF
    __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_ON & _STVREN_ON & _BORV_LO & _LVP_OFF

;*************************  Register Definition  *********************
			UDATA_SHR
flowcounter			res		1	;program flow counter for interrupts.
f_resultL			res		1	;fereq measurement value(duration of first 20 sine wave periods)
f_resultH			res		1	;
t_resultL			res		1	;decay time measurement value
t_resultH			res		1	;
periodcount			res		1	;stores the number of detected oscillations, after every pulse
periodresult		res		1
tempL				res		1
tempH				res		1
flags				res		1	;various flags for communication between interrupts and main, and between main and sub
beeptemp			res		1	;used in 0.5ms delay and 2ms delay
FetPulseTimeRam		res		1	;ram copy of constants for using in timer setting
PauseTimeRam		res		1
debounce_counter	res		1	;debounce timer counter
temp1				res		1	
double_p_temp		res		1	;used in double push detector

#define	soundflag	flags,0
#define	soundpol	flags,2
#define	dataready	flags,3
#define	datareadyF	flags,4
#define	serialmode	flags,5

buf1			UDATA		0x20
Counter1			res		1	;used in adaptive reference calculation
sum1_LSB			res		1	;used in adaptive reference calculation
sum2				res		1	;used in adaptive reference calculation
sum3				res		1	;used in adaptive reference calculation
sum4				res		1
ReferenceL			res		1	;result of adaptive reference calculation
ReferenceH			res		1
IndexW				res		1	;index for writing to circular buffer
IndexR				res		1	;index for reading from circular buffer
CircularBuffer		res		64
							;74 bytes total block

moving_avg		UDATA		0xA0	;second circular buffer 16 2Byte samples, used in moving average calculation for t_result
IndexW2				res		1	;index for writing to circular buffer
IndexR2				res		1	;index for reading from circular buffer
Addtmp				res		1	;extra byte for adding
CircularBuffer2		res		32	;16x2byte

delay1		UDATA
temp2			res		1	;used in 250mS delay
temp3			res		1	;used in 250ms delay
temp4			res		1	;used in startup beep counter

serial		UDATA
TxPointer		res		1	;pointer to data in buffer
txdelay			res		1	;transmit update rate counter

bin2ASCII		UDATA
BINL            res     .1                      ; Binary value to convert
BINH            res     .1
BCDL            res     .1                      ; Resulting BCD value
BCDH            res     .1
BCDU            res     .1
COUNT           res     .1                      ; Counter
ASCII			res		.6						;resulting string 5bytes msb first

moving_avg_3		UDATA			;third circular buffer 16 2Byte samples, used in moving average calculation for t_result
IndexW2_3				res		1	;index for writing to circular buffer
IndexR2_3				res		1	;index for reading from circular buffer
Addtmp_3				res		1	;extra byte for adding
CircularBuffer2_3		res		32	;16x2byte

;************************* Output devices Definition   *********************
MOSFET			EQU		RA5		;MOSFET drive connection port	(active high)
LED1			EQU		RA1		;Led contact
Piezo1			EQU		RA0		;piezo buzzer contacts
Piezo2			EQU		RA2
MEASURING_INPUT	EQU		RA4		;Measuring input. DO NOT change! Changing it here will not change comparator input port.
rs232_tx_out	EQU		RA0		;serial output(no not change! it is hardware TX)
;************************* Input devices Definition   *********************
push_button 	EQU		RA3		;Button connection

;************************* Constants Definition   *********************
FetPulseTime	EQU		.136  	;MOSFET pulse width in microseconds		(range  20-255)(actual value=n+4us)
PauseTime		EQU		.255	;n*8=[us] 230*8=1840us  (range 2-255)
				;total	 2000us = 2ms period
sensitivity		EQU		.15		;min difference between reference and current value that causes "beep"
CompRef			EQU		.15		;n for comparator reference voltage:Vref=n*1,024/32(range 0-31))

SPBRGL_VAL		EQU		.51		;set serial baud rate 9600 for 32Mhz clock
;button constants
shortpush_max		EQU		.25			;short push maximum duraton*20 =[ms], longer than this is long push
double_push_min		EQU		.1			;min wait between double push pushes [ms]/20, button release debounce
double_push_max		EQU		.25			;max wait between pulses [ms]/20 ,if longer than this pushes count as separate single pulses.
;****************************************************************
;Beginning of program
;****************************************************************
RESET_VECT  CODE    0x0000
    	PAGESEL START
    	GOTO    START

;------------------------------------------------------------------------------
; INTERRUPT SERVICE ROUTINE
;------------------------------------------------------------------------------

ISR		CODE	0x0004
	
		btfss	INTCON,TMR0IF		;ewerything past this point is for timer0 interrupt only
		goto	OtherInt				;just in case
		btfsc	flowcounter,0
		goto	Start_pulse 		;if flowcounter bit0, happens this
		btfsc	flowcounter,1
		goto	Stop_pulse			;if flowcounter bit1, happens this
OtherInt
		banksel	PIR2
		btfsc	PIR2,C1IF
		goto	ComparatorInterrupt
		btfsc	PIR1,TMR2IF
		goto	Timer2Interrupt
		btfsc	INTCON,IOCIF
		goto	IntOnChange			;to serial data
		
		goto	error1				;catch false interrupts
		retfie	;never reach this

; INTERRUPT SERVICE ROUTINE SUBROUTINES
		CODE

Start_pulse						;Starts mosfet pulse
        movlw   b'11000010'     ; configure Timer0:
                ; --0-----          timer mode (T0CS = 0), internal instruction clock(8 MHz)
                ; ----0---          prescaler assigned to Timer0 (PSA = 0)
                ; -----010          prescale = 8
        banksel OPTION_REG      ;   -> increment TMR0 every 1 us
        movwf   OPTION_REG				;delay with timer0        	
		comf	FetPulseTimeRam,w		;Nii saame ige vrtuse lespoole loendava taimeri jaoks, see kohe W-sse ka		
		banksel TMR0
		movwf	TMR0			;lad timer with FetPulseTime, starting timer
		bcf		INTCON,TMR0IF	;clear timer0 interrupt flag
		;---------------------------------------------
		banksel T1CON			;
		bcf		T1CON,TMR1ON			;timer1 seisma
		bcf		T1GCON,TMR1GE
		;-----------------------save result (from previous pulse)-----------------------------
		btfsc	dataready		;test if data was successfully copied, if not then skip update.
		goto	skp_upd
		banksel TMR1L
		movf	TMR1L,w
;		movlw	.54			;test
		movwf	t_resultL
		movf	TMR1H,w
;		movlw	.3			;test
		movwf	t_resultH
		movf	periodcount,w
		banksel	periodresult
		movwf	periodresult	
		bsf		dataready		;set data ready flag			
skp_upd

		;-----------------------clear timer1---------------------------
		banksel	TMR1L
		clrf	TMR1L
		clrf	TMR1H		
		;----------------------------------------------------------------
		banksel	LATA
		bsf		LATA,MOSFET		;Turn on mosfet
		clrf	flowcounter
		bsf		flowcounter,1	;next step nr1,  Stop_pulse, set up for next timer0 interrupt
		;-----------------------comparator interrupt setup------------
;		bcf		INTCON,PEIE		;
		banksel	PIE2
		bcf		PIE2,C1IE		;comparator interrupt disable
		banksel	PIR2
		bcf		PIR2,C1IF		;clear comparator interrupt flag	
		;--------------------------------------------------------------
	retfie


Stop_pulse		
		;----------------------------------------------
		banksel	TMR0
		btfss	TMR0,2			;let timer0 count extra 4us, to compensate out possible other (sound) interrupts
		bra		$-1
		banksel LATA
		bcf		LATA,MOSFET		;Mosfet kinni
		;------------------------mosfet turnoff delay--------
		call	rtrn2			;1us delay
		call	rtrn2			;1us delay
		;------------------------timer1 kima ---------------------
		banksel T1CON
		movlw	b'01000001'		;
				;'01------'		timer1 clock source is system clock 32MHz
				;'--00----'		prescale off(1:1)
				;'----xxx-'		ebaolulised
				;'-------1'		timer1 enable TMR1ON
		movwf	T1CON			;seadistatud
		;-----------------------comparator interrupt setup------------
		bsf		INTCON,PEIE		;
		banksel	PIE2
		bsf		PIE2,C1IE		;comparator interrupt enable
		banksel	PIR2
		bcf		PIR2,C1IF		;clear comparator interrupt flag			
		;-----------------------timer0 setup----------
		clrf	flowcounter
		bsf		flowcounter,0	;jrgmine samm nr0, ehk Start_Pulse
		movlw   b'11000101'     ; configure Timer0:
                ; --0-----          timer mode (T0CS = 0), sisemine instruktsiooni kell(8 MHz)
                ; ----0---          prescaler assigned to Timer0 (PSA = 0)
                ; -----101          prescale = 64
        banksel OPTION_REG      ;   -> increment TMR0 every 8 us
        movwf   OPTION_REG	
		comf	PauseTimeRam,w		;Nii saame ige vrtuse lespoole loendava taimeri jaoks, see kohe W-sse ka		
		banksel TMR0
		movwf	TMR0			;laadime taimerisse PauseTime, seega taimer kima
		bcf		INTCON,TMR0IF	;clear timer0 interrupt flag,
		;----------------------counter------------------	
		clrf	periodcount

	retfie

ComparatorInterrupt
		banksel	PIR2
		bcf		PIR2,C1IF		;clear comparator interrupt flag
		;-------------------
		incf	periodcount,f
		;------------------
		movlw	.19
		subwf	periodcount,w	;check for 19-periodcount
		btfsc	STATUS,Z
		goto	cycle19

		movlw	.20
		subwf	periodcount,w	;check for 20-periodcount
		btfsc	STATUS,Z
		goto	cycle20

	retfie
cycle19		;----------------activate timer1 gate function so that timer stops at next rising edge
		banksel T1GCON
		bsf		T1GCON,T1GPOL
		movlw	b'01111010'		;
				;'0-------'		;gate function disabled TMR1GE
				;'-1------'		;gate active high
				;'--11----'		;gate single pulse and toggle mode
				;'------10'		;gate input from comparator
				;'----1---'		;t1ggo
		movwf	T1GCON			;
		bcf		T1GCON,T1GPOL
		bsf		T1GCON,T1GPOL	;force start gate circuit
		bsf		T1GCON,TMR1GE	;switch to gate control
	retfie
cycle20
		;-----------------------save result, this is duration of first 20 periods.-----------------------------
		btfsc	datareadyF		;test if data was successfully copied, if not then skip update.
		goto	skp_upd2
		banksel TMR1L
		movf	TMR1L,w
		movwf	f_resultL
		movf	TMR1H,w
		movwf	f_resultH
		bsf		datareadyF		;set data ready flag
skp_upd2
		;-----------------------clear timer1---------------------------
		clrf	TMR1L
		clrf	TMR1H	
		;--------------------start timer1, new settings---------------
		banksel T1GCON
		movlw	b'10000010'		;
				;'1-------'		;gate function enabled
				;'-0------'		;gate active low
				;'------10'		;gate input from comparator
		movwf	T1GCON			;
		banksel T1CON
		movlw	b'01000001'		;
				;'01------'		timer1 clock source is system clock 32MHz
				;'--00----'		prescale off(1:1)
				;'----xxx-'		ebaolulised
				;'-------1'		timer1 enable TMR1ON
		movwf	T1CON			;seadistatud
		;-------------------disable comparator interrupt--------------------
;		bcf		INTCON,PEIE		
	retfie

;-------------------------Timer2 interrupt, sound generation--------------
Timer2Interrupt			;
		banksel	PIR1
		bcf		PIR1,TMR2IF		;clear timer2 interrupt flag

		btfss	soundpol	
		goto	neg
pos
		bcf		soundpol
		banksel	LATA
		bcf		LATA,Piezo1
		btfsc	soundflag
		bsf		LATA,Piezo2
		retfie
neg
		bsf		soundpol
		banksel	LATA
		bcf		LATA,Piezo2
		btfsc	soundflag
		bsf		LATA,Piezo1
		retfie

;----------------------------Interrupt on change, invert other pin voltage for RS-232 sending, then we do not need level converter ---------------------------
IntOnChange			;--------RS-232 lines connected to RP0 and RP2
		banksel PORTA
		btfss	PORTA,rs232_tx_out
		goto	invbit
		banksel LATA
		bcf		LATA,Piezo2
		banksel	IOCAF
		movlw	0x00
		ANDWF	IOCAF, F		;clear interrupt on change flags
		retfie
invbit
		banksel LATA
		bsf		LATA,Piezo2
		banksel	IOCAF
		movlw	0x00
		ANDWF	IOCAF, F		;clear interrupt on change flags
		retfie

;***************************************************************************************
;
; MAIN PROGRAM
;
;***************************************************************************************
PGM		CODE		
START	
		call	_2ms
		banksel OSCCON
		movlw	b'11110000'
		movwf	OSCCON			;configure oscillator to internal 32MHz
		banksel PORTA
		clrf	PORTA			;init portA
		banksel LATA
		clrf	LATA			;init port data latch
		banksel ANSELA
		clrf	ANSELA			;all digital I/O
		bsf		ANSELA,MEASURING_INPUT	;set comparator input analog
 		banksel TRISA
      	movlw	b'00111111'		;Set TRISA port: 0 =output, 1=input, all to input first
		movwf	TRISA	   	    ;
		banksel LATA
		bcf		LATA,MOSFET	
		banksel	TRISA
		bcf		TRISA,MOSFET	;MOSFETi to output
		bcf		TRISA,Piezo1	;piezo to output
		bcf		TRISA,Piezo2
		bcf		TRISA,LED1		;led ots vljundiks
		banksel	WPUA
		bsf		WPUA,WPUA3		;enable weak pull-up		
		;------------------- init variables ------------------------------------
		clrf	flowcounter		

		movlw	FetPulseTime
		movwf	FetPulseTimeRam

		movlw	PauseTime
		movwf	PauseTimeRam

		banksel	Counter1
		clrf	Counter1
		clrf	IndexW
		clrf	ReferenceL
		clrf	ReferenceH
		clrf	sum1_LSB
		clrf	sum2
		clrf	sum3	

		banksel	IndexW2
		clrf	IndexW2
		;----------------------DAC setup-------------------------------------
		banksel	DACCON0			;
		movlw	b'11000000'		;positive ref Vdd
		movwf	DACCON0			;
		movlw	CompRef			;ref voltage for comparator, defined in constants section
		movwf	DACCON1			;
		;---------------------setup comparator-------------------------------
		banksel	CM1CON0
		movlw	b'10010101'		;
				;'1-------'		;comparator enable
				;'--0-----'		;comparator output internal only
				;'---1----'		;output inverted
				;'------0-'		;hysteresis disabled
				;'-------1'		;sync with timer1
		movwf	CM1CON0			;
		movlw	b'10010001'		;
				;'1-------'		;rising edge interrupt
				;'--01----'		;+input connects to internal DAC Voltage reference
				;'-------1'		;-input connects to C1IN1- (RA4) 
		movwf	CM1CON1			;
		;--------------------------------------	
		bsf		flowcounter,0	;jrgmine samm nr0, ehk Start_Pulse
		banksel INTCON			;enable interrupts
		movlw 	b'10100000'		
				; 1-------		general interrupt enable
				; --1-----		timer0 interrupt enable
		movwf	INTCON
		;----------------------------------------------------------------------
		bsf		flowcounter,0	;next step nr0, ehk Start_Pulse
		bsf		INTCON,TMR0IF	;begin sending pulses, generate interrupt with soft, hakkame impulsse saatma, selleks tekitame kunstliku interrupti

		;--------------------alive led-------------
		banksel LATA
		bsf		LATA,LED1
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		banksel LATA
		bcf		LATA,LED1
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		banksel LATA
		bsf		LATA,LED1
		;-----------------------config sound-------------------
		call	confTimer2
		;-----------------------startup beep--------------------
		bsf		soundflag		;sound on
		call	_250mS
		call	_250mS
		bcf		soundflag		;sound off
		;---------------init buffer-----------------------------
		call	init_buffer
		;--------------init some other things-------------------
		banksel	RCSTA
		clrf	RCSTA			;disable serial port
		banksel	IOCAN
		bcf		IOCAN,rs232_tx_out	;disable negative going edge interrupt on RA0
		banksel	IOCAP
		bcf		IOCAP,rs232_tx_out	;disable positive going edge interrupt on RA0
		banksel IOCAF
		movlw	0x00
		ANDWF	IOCAF, F		;clear all interrupt on change flags
		bcf		INTCON,IOCIE	;disable interrupts on change

		bcf		serialmode
;****************************************************************
;* MAIN                                                         *
;****************************************************************
Main	;main loop
		btfss	dataready		;wait for data to become ready
		goto	bpress			;if data not ready skip calculations and serial

		btfss	datareadyF		;check for data to be ready
		goto	bpress			;if data not ready skip calculations and serial

		btfsc	serialmode		;check if serial communication is activated
		goto 	serial			;if serial mode then skip calculations
		;--------------------------main calculations----------------------
		call	moving_average_3	;filter f_result
		call	nonferro			;nonferro block, free to comment out if not needed.
		call	calc
		call	fast_reference		;moving mode
		bcf		dataready
		bcf		datareadyF
		goto 	bpress
		;---------------------------------------------------------------
serial
		call	moving_average
		call	moving_average_3
		banksel	txdelay
		incf	txdelay,f		;do not send so often
		btfsc	STATUS,Z
		call	sendserial_main
		bcf		dataready
		bcf		datareadyF
bpress
		banksel	PORTA		
		btfss	PORTA,push_button		;detect if button is pressed
		call	button			;this button routine takes time, some mearsurements are lost while.
		goto Main

;****************************************************************
;* SUBROUTINES 			*
;****************************************************************	
confTimer2		;--------------------config Timer2 and interrupts, for sound---------
		;start making sound
		banksel PR2
		movlw	0xFF
		movwf	PR2
		banksel	T2CON
		movlw	b'00000110'		;timer2 on, postscaler 1:1, prescaler 1:16
		movwf	T2CON			;timer interval 0,5ms
		clrf	TMR2
		banksel	PIR1
		bcf		PIR1,TMR2IF		;clear timer2 interrupt flag
		banksel	PIE1
		bsf		PIE1,TMR2IE		;enable timer2 interrupts
		bsf		INTCON,PEIE		;periferial interrupt enable
		return
confTimer2disable
		banksel	T2CON
		clrf	T2CON			;disable timer2
		banksel	PIR1
		bcf		PIR1,TMR2IF		;clear timer2 interrupt flag
		banksel	PIE1
		bcf		PIE1,TMR2IE		;disable timer2 interrupts	
		return			
sendserial_main
		;-------------start sending serial-------------
		;----send sagedus
		banksel	BINL
		movf	f_resultL,w
		movwf	BINL
		movf	f_resultH,w
		movwf	BINH
		call	Bin2BCD
		call	BCD2ASCII
		TransmitSerial_FlashString sagedus_table ;macro from metalmacros.inc
		call	TransmitSerial_MemString
		;---send sumbumine-----
		banksel	BINL
		movf	t_resultL,w
		movwf	BINL
		movf	t_resultH,w
		movwf	BINH
		call	Bin2BCD
		call	BCD2ASCII
		TransmitSerial_FlashString sumbumis_table ;macro from metalmacros.inc
		call	TransmitSerial_MemString
		;----send period count
		banksel	BINL
		movf	periodresult,w
		movwf	BINL
		clrf	BINH
		call	Bin2BCD
		call	BCD2ASCII
		TransmitSerial_FlashString periood_table ;macro from metalmacros.inc
		call	TransmitSerial_MemString
		;------------------
		return
sagedus_table
		brw
		dt 0x0A,0x0D,"freq_decay_count;",0x00  ;0x0A=line feed, 0x0D=carriage return, 0x00=string end,
			;0x0C=clear screen
sumbumis_table
		brw
		dt 	";",0x00  ;0x0A=line feed, 0x0D=carriage return, 0x00=string end,
			;0x0C=clear screen
periood_table
		brw
		dt 	";",0x00  ;0x0A=line feed, 0x0D=carriage return, 0x00=string end,
			;0x0C=clear screen
		;-----------------Set up serial port and buffers.-----------------
SetupSerial
		banksel	TRISA						
		bcf		TRISA,rs232_tx_out ;set tris bit for TX 
		banksel	ANSELA
		bcf		ANSELA,rs232_tx_out

		banksel APFCON
		bcf		APFCON,TXCKSEL	;set TX funcion to RA0

		banksel SPBRGL
		movlw	SPBRGL_VAL	;set baud rate
		movwf	SPBRGL

		banksel	TXSTA
		movlw	b'00100000'		;enable transmission and low baud rate
		movwf	TXSTA

		banksel	RCSTA
		movlw	b'10000000'		;enable serial port and disable reception
		movwf	RCSTA
		return
;-------------------TransmitSerial_MemString---------------------
;	Transmits null terminated string from RAM
;	input:string starting address ASCII 
;	uses: none
;----------------------------------------------------------------
TransmitSerial_MemString
		movlw	high ASCII
		movwf	FSR0H
		movlw	low ASCII
		movwf	FSR0L
Txpoint_ms
		movf	INDF0,W
		btfsc	STATUS,Z	;test for table end
		return				;operation complete	
		banksel PIR1
		btfss	PIR1,TXIF	;wait for transmitter to be ready
		goto	$-.1		;
		banksel	TXREG
		movwf	TXREG		;and transmit
		addfsr	FSR0,1		;prepare next address for reading, add 1 to file select register address
		goto	Txpoint_ms


		;---------------------------16 bit Binary to 5digit BCD format
		;input:	BINL:BINH	(input is destroyed)
		;BitWise's routine returns 5 digits in packed BCD format (BCDU:BCDH:BCDL)
 		;-------------------------------------------------------------------------------
Bin2BCD
				banksel	BCDL
                 clrf    BCDL                    ; Clear result
                 clrf    BCDH
                 clrf    BCDU
                 movlw   .16                     ; Set bit counter
                 movwf   COUNT
ConvertBit
                 movlw   h'33'                   ; Correct BCD value so that
                 addwf   BCDL,F                  ; subsequent shift yields
                 btfsc   BCDL,.3                 ; correct value.
                 andlw   h'f0'
                 btfsc   BCDL,.7
                 andlw   h'0f'
                 subwf   BCDL,F
 
                 movlw   h'33'
                 addwf   BCDH,F
                 btfsc   BCDH,.3
                 andlw   h'f0'
                 btfsc   BCDH,.7
                 andlw   h'0f'
                 subwf   BCDH,F
 
                 rlf     BINL,F                  ; Shift out a binary bit
                 rlf     BINH,F
 
                 rlf     BCDL,F                  ; .. and into BCD value
                 rlf     BCDH,F
                 rlf     BCDU,F
 
                 decfsz  COUNT,F                 ; Repeat for all bits
                 bra     ConvertBit
                 return     
;-----------------------5 digits BCD (BCDU:BCDH:BCDL) to 5byte ASCII Msb first
BCD2ASCII
		banksel	BCDU
		movf	BCDU,w						;digit1
		andlw	B'00001111'
		iorlw	0x30
		movwf	ASCII

		swapf	BCDH,w						;digit2				
		andlw	B'00001111'
		iorlw	0x30
		movwf	ASCII+.1

		movf	BCDH,w						;digit3
		andlw	B'00001111'
		iorlw	0x30
		movwf	ASCII+.2

		swapf	BCDL,w						;digit4				
		andlw	B'00001111'
		iorlw	0x30
		movwf	ASCII+.3

		movfw	BCDL						;digit5
		andlw	B'00001111'
		iorlw	0x30
		movwf	ASCII+.4
		
		clrf	ASCII+.5					;NULL terminator

	return

	;-----------------------------------------------------------------	
	; Non ferro detecting and ground compensation subroutine.
	;	add 0x9C00 (39936) to f_result
	;	complement f_result,
	;	add with t_result, and store in t_result
	;	RAM usage
	;Input: f_resultL, f_resultH, t_resultL, t_resultH
	;output: t_resultL, t_resultH
	;Uses: none
	;-----------------------------------------------------------------
nonferro
		movlw	0x9C
		addwf	f_resultH,f	

		comf	f_resultL,w	
		addwf	t_resultL,f
		comf	f_resultH,w
		addwfc	t_resultH,f
		return

	;-----------------------------------------------------------------	
	;init buffer block. Clear all buffer contetns.
	;-----------------------------------------------------------------
init_buffer
		clrf	IndexR
		movlw	high CircularBuffer
		movwf	FSR0H
		movlw	low CircularBuffer	;
		movwf	FSR0L			;prepare first address for reading	
init_loop
		clrw
		movwi	FSR0++
		incf	IndexR,f		;next address
		movwi	FSR0++
		incf	IndexR,f		;next address
		btfss	IndexR,6		;if bit 6 is set, it has counted 64 times then exit loop
		goto	init_loop			;
		clrf	IndexR
		banksel sum1_LSB
		clrf	sum1_LSB
		clrf	sum2
		clrf	sum3
	return

	;------------------------------------------------------------------------------------
	;	ver2	Fast adaptive reference block for Moving Mode.
	;	Updates reference after every 16 pulses(32 msec) with maximum element of last 32 averaged sample blocks(1sek)(16samples average each block)
	;														
	;------------------------------------------------------------------------------------	
fast_reference
		;----------------------- sum = sum + new result -------------------------------
		banksel sum1_LSB		
		movf	t_resultL,w	
		addwf	sum1_LSB,f
		movf	t_resultH,w
		addwfc	sum2,f
		clrw
		addwfc	sum3,f
		;------------------------------------------------------------------------------
		banksel	Counter1
		incf	Counter1,f
		btfss	Counter1,4		;bit 4 sets when counting reaches 16,decide update buffer or
		return					;not yet to update buffer
		clrf	Counter1
		;------------------------divide by 16, shift right 4 times----------------------
		lsrf	sum3,f			
		rrf		sum2,f
		rrf		sum1_LSB,f

		lsrf	sum3,f
		rrf		sum2,f
		rrf		sum1_LSB,f	

		lsrf	sum3,f
		rrf		sum2,f
		rrf		sum1_LSB,f	

		lsrf	sum3,f
		rrf		sum2,f
		rrf		sum1_LSB,f		
		;-----------------------update ring buffer-------------------------------------	
		banksel CircularBuffer
		movlw	high CircularBuffer
		movwf	FSR0H
		movlw	low CircularBuffer	;
		addwf	IndexW,W		;
		movwf	FSR0L			;prepare address for writing
		movf	sum1_LSB,w			;
		movwf	INDF0			;write LSB value to buffer (little endian)

		incf	IndexW,f		;next address in buffer

		movf	sum2,w			;			
		movwi	++FSR0			;write MSB value to buffer and pre inc FSR0L
		
		incf	IndexW,f		;next address in buffer

		btfsc	IndexW,6		;if bit 6 is set, it has counted 64 times(32 element buffer)
		clrf	IndexW			;then reset index

		clrf	sum1_LSB		;do not need those any more
		clrf	sum2			;
		clrf	sum3			;
		clrf	sum4
		;--------------ring buffer reading loop, detect maximum value in buffer
		clrf	IndexR

		movlw	low CircularBuffer	;
		movwf	FSR0L			;prepare address for reading
rloop2
		moviw	FSR0++			;read LSB value from buffer
		movwf	sum3			;

		incf	IndexR,f		;next address

		moviw	FSR0++			;read MSB value from buffer
		movwf	sum4			;add to sum

		incf	IndexR,f		;next address
						;--------detect biggest elemant--------------
						;--------compare old value "sum1:sum2" new value "sum3:sum4"------
		movf	sum3,w			
		subwf	sum1_LSB,w		;sub sum1-sum3
		movf	sum4,w			;compare old value to new
		subwfb	sum2,w
		btfsc	STATUS,C		;if new value>=current maximum then copy new value
		goto	skipcopy		;else skip copy	
		movf	sum3,w
		movwf	sum1_LSB
		movf	sum4,w		
		movwf	sum2
skipcopy		
		btfss	IndexR,6		;if bit 6 is set, it has counted 64 times(32 element buffer) then exit loop
		goto	rloop2			;
		clrf	IndexR
				;						new reference now in sum2:sum1_LSB
		;----------------------------------------------------------
								;update reference		
		movf	sum1_LSB,w			
		movwf	ReferenceL
		movf	sum2,w
		movwf	ReferenceH

		clrf	sum1_LSB		;do not need those any more
		clrf	sum2			;
		clrf	sum3			;
		clrf	sum4
	return


	;------------------------------------------------------------------------------------
	;		Moving average block. Uses circular buffer to calculate average of last 16 samples
	;	Input:t_resultL,t_resultH
	;	Output:t_resultL,t_resultH(updated)
	;	uses:moving_average(mem block)
	;------------------------------------------------------------------------------------	
moving_average		;real time moving average of 16 measurements, simple non recursive algorithm
		banksel CircularBuffer2

		;---------------writing section------------------
		movlw	high	CircularBuffer2
		movwf	FSR0H
		movlw	low		CircularBuffer2	;
		addwf	IndexW2,W		;
		movwf	FSR0L			;prepare address for writing
		movf	t_resultL,w			;
		movwf	INDF0			;write LSB value to buffer (little endian)

		incf	IndexW2,f		;next address in buffer

		movf	t_resultH,w			;
		movwi	++FSR0			;write MSB value to buffer
		
		incf	IndexW2,f		;next address in buffer

		btfsc	IndexW2,5		;if bit 5 is set, it has counted 32
		clrf	IndexW2			;then reset index

		;--------------- reading section-------------------
		clrf	t_resultL
		clrf	t_resultH
		clrf	Addtmp
		clrf	IndexR2
read_avg_loop
		movlw	CircularBuffer2	;
		addwf	IndexR2,W		;
		movwf	FSR0L			;prepare address for reading
		movf	INDF0,W			;read LSB value from buffer
		addwf	t_resultL,f		;add to output

		incf	IndexR2,f		;next address

;		addfsr	FSR0,1			;prepare address for reading, add 1 to file select register address
		moviw	++FSR0			;read MSB value from buffer
		addwfc	t_resultH,f		;add to output
		clrw
		addwfc	Addtmp,f

		incf	IndexR2,f		;next address

		btfss	IndexR2,5		;if bit 5 is set, it has counted 32 then exit loop
		goto	read_avg_loop	;

		lsrf	Addtmp,f			;divide by 16
		rrf		t_resultH,f	
		rrf		t_resultL,f	

		lsrf	Addtmp,f			
		rrf		t_resultH,f	
		rrf		t_resultL,f

		lsrf	Addtmp,f			
		rrf		t_resultH,f	
		rrf		t_resultL,f	

		lsrf	Addtmp,f			
		rrf		t_resultH,f	
		rrf		t_resultL,f		;ready values in t_resultL and H

	return

	;------------------------------------------------------------------------------------
	;		Moving average block 3. Uses circular buffer to calculate average of last 16 samples
	;	Input:f_resultL,f_resultH
	;	Output:f_resultL,f_resultH(updated)
	;	uses:moving_average_3(mem block)
	;------------------------------------------------------------------------------------	
moving_average_3		;real time moving average of 16 measurements, simple non recursive algorithm
		banksel CircularBuffer2

		;---------------writing section------------------
		movlw	high	CircularBuffer2_3
		movwf	FSR0H
		movlw	low		CircularBuffer2_3	;
		addwf	IndexW2_3,W		;
		movwf	FSR0L			;prepare address for writing
		movf	f_resultL,w			;
		movwf	INDF0			;write LSB value to buffer (little endian)

		incf	IndexW2_3,f		;next address in buffer

		movf	f_resultH,w			;
		movwi	++FSR0			;write MSB value to buffer
		
		incf	IndexW2_3,f		;next address in buffer

		btfsc	IndexW2_3,5		;if bit 5 is set, it has counted 32
		clrf	IndexW2_3			;then reset index

		;--------------- reading section-------------------
		clrf	f_resultL
		clrf	f_resultH
		clrf	Addtmp_3
		clrf	IndexR2_3
read_avg_loop_3
		movlw	CircularBuffer2_3	;
		addwf	IndexR2_3,W		;
		movwf	FSR0L			;prepare address for reading
		movf	INDF0,W			;read LSB value from buffer
		addwf	f_resultL,f		;add to output

		incf	IndexR2_3,f		;next address

;		addfsr	FSR0,1			;prepare address for reading, add 1 to file select register address
		moviw	++FSR0			;read MSB value from buffer
		addwfc	f_resultH,f		;add to output
		clrw
		addwfc	Addtmp_3,f

		incf	IndexR2_3,f		;next address

		btfss	IndexR2_3,5		;if bit 5 is set, it has counted 32 then exit loop
		goto	read_avg_loop_3	;

		lsrf	Addtmp_3,f			;divide by 16
		rrf		f_resultH,f	
		rrf		f_resultL,f	

		lsrf	Addtmp_3,f			
		rrf		f_resultH,f	
		rrf		f_resultL,f

		lsrf	Addtmp_3,f			
		rrf		f_resultH,f	
		rrf		f_resultL,f	

		lsrf	Addtmp_3,f			
		rrf		f_resultH,f	
		rrf		f_resultL,f		;ready values in t_resultL and H

	return
;----------------------------------------------------------------------------------------
calc							;compare current result and reference
		movf	t_resultL,w		;copy result to temp
		movwf	tempL
		movf	t_resultH,W
		movwf	tempH

		movlw	sensitivity				;add sensitivity
		addwf	tempL,f
		clrw
		addwfc	tempH,f

		banksel	ReferenceL				;compare
		movf	tempL,w
		subwf	ReferenceL,w
		movf	tempH,w
		subwfb	ReferenceH,w
		bcf		soundflag			;sound off or
		btfsc	STATUS,C
		bsf		soundflag			;sound on

		return
;----------------------------------------------------------------------------------------

error1				;low tone to indicate that error happened
		call	_2ms
		call	_2ms
		banksel	LATA
		bcf		LATA,Piezo1
		bsf		LATA,Piezo2
		call	_2ms
		call	_2ms
		banksel	LATA
		bcf		LATA,Piezo2
		bsf		LATA,Piezo1
	goto	error1

;---------------------------------------------------------------------------------------
;		button press detection, when holding button down for 0,75 turns power off(Sleep)
;--------------------------------------------------------------------------------------
button					;if button is pressed for long goto sleep, if double push activate serial mode
		;-----------------------------push type detection start----------------------------------
wait_release	;determine is it long press or short
		movlw	shortpush_max		;load wait time
		movwf	temp1
loop1
		banksel	PORTA
		btfsc	PORTA,push_button		;if button released go to double push detection
		goto	double_p
		call	_20ms
		decfsz	temp1,f
		goto	loop1
		goto 	long_push


double_p								;detect if this is double push
		call	_20ms					;debounce button release
		movlw	double_push_max-.1			;minus debounce already waited
		movwf	double_p_temp
loop2
		call	_20ms
		banksel	PORTA
		btfss	PORTA,push_button
		goto	double_push
		decfsz	double_p_temp,f
		goto	loop2
		goto	short_push			;if second push did not come, stick with short push
		;-----------------------------push type detection end----------------------------------

long_push
		banksel LATA
		bcf		LATA,LED1
		banksel	PORTA
		btfss	PORTA,push_button		;wait for button to be released
		goto	$-.1

sleepy  		;----------------------prepare for going to sleep-----------
		bcf		INTCON,GIE		;disable all interrupts, device will still wake from sleep
		bcf		INTCON,PEIE		;disable periferial interrupts
		bcf		INTCON,TMR0IE	;disable Timer0 interrupts

		banksel LATA
		bcf		LATA,MOSFET		;turn off mosfet before going to sleep
		bcf		LATA,Piezo1
		bcf		LATA,Piezo2

		call	_250mS			;debounce button release
		banksel	CM1CON0
		bcf		CM1CON0,C1ON	;disable comparator to save power
		banksel	DACCON0
		bcf		DACCON0,DACEN	;disable DAC voltage ref to save power
		;---------------------interrupt on change config for wake up from sleep----------------------
		banksel	IOCAN
		bsf		IOCAN,IOCAN3	;enable negative going edge interrupt on RA3
;		MOVLW	0xff			;required sequence to clear interrupt on change flags
;		XORWF	IOCAF, W		;
		movlw	0x00
		ANDWF	IOCAF, F		;clear interrupt on change flags
		bsf		INTCON,IOCIE	;enable interrupts on change		
		SLEEP
		;------------------------------SLEEP------------------------------------------------------
		nop
		bcf		INTCON,IOCIE	;disable interrupts on change
		banksel	IOCAF
		movlw	0x00
		ANDWF	IOCAF, F		;clear interrupt on change flags
		banksel	CM1CON0
		bsf		CM1CON0,C1ON	;enable comparator
		banksel	DACCON0
		bsf		DACCON0,DACEN	;enable DAC voltage ref
		call	init_buffer		;clear buffer
		call	_2ms		
		bsf		INTCON,PEIE		;enable periferial interrupts
		bsf		INTCON,TMR0IE	;enable Timer0 interrupts
		bsf		INTCON,GIE		;general interrupt enable
		call	_2ms
		call	_2ms
		banksel LATA
		bsf		LATA,LED1
		return
double_push		;-------------------activate serial transmit mode------------------------
		call	confTimer2disable
		bsf		serialmode
		call	SetupSerial
		;------------------invert other output pin with interrupt on change
		banksel	IOCAN
		bsf		IOCAN,rs232_tx_out	;enable negative going edge interrupt on RA0
		banksel	IOCAP
		bsf		IOCAP,rs232_tx_out	;enable positive going edge interrupt on RA0
		banksel IOCAF
;		MOVLW	0xff			;required sequence to clear interrupt on change flags
;		XORWF	IOCAF, W		;
		movlw	0x00
		ANDWF	IOCAF, F		;clear all interrupt on change flags
		bsf		INTCON,IOCIE	;enable interrupts on change			
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		call	_250mS
		return
short_push		;--------------------go back to normal mode----------------------
		bcf		serialmode
		banksel	RCSTA
		clrf	RCSTA			;disable serial port
		banksel	IOCAN
		bcf		IOCAN,rs232_tx_out	;disable negative going edge interrupt on RA0
		banksel	IOCAP
		bcf		IOCAP,rs232_tx_out	;disable positive going edge interrupt on RA0
		banksel IOCAF
		movlw	0x00
		ANDWF	IOCAF, F		;clear all interrupt on change flags
		bcf		INTCON,IOCIE	;disable interrupts on change	
		banksel	TRISA
		bcf		TRISA,Piezo1	;piezo to output
		bcf		TRISA,Piezo2
		call	confTimer2		;reconfigure timer2
		return
		
;****************************************************************
;* Delays 			*
;****************************************************************
rtrn						;0,5 microsecond delay
	return
rtrn2						;1 microsecond delay
		goto	$+1
		goto	$+1
	return
_1ms
		movlw	.245	;2
		movwf	beeptemp	;2
looptm1	call	rtrn	; siit maalt korratakse
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn	
		decfsz	beeptemp,f	
		goto	looptm1
		return
_20ms						;20ms delay(for 32Mhz clock) debuggeris testitud
		banksel temp3
		movlw	.40			;fine tuning
		movwf	temp3
		movlw	.21			;rough tuning
		movwf	temp2
D20		call	rtrn2	
		call	rtrn2
		call	rtrn2
		call	rtrn	
		decfsz 	temp3,f
		goto 	D20
		decfsz 	temp2,f
		goto 	D20		
	return

_2ms
		movlw	.243	;2
		movwf	beeptemp	;2
looptm	call	rtrn	; siit maalt korratakse
		call	rtrn	
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn
		call	rtrn	;callisid 60 tskli jagu
		decfsz	beeptemp,f	
		goto	looptm
		return


_250mS						;250mS second delay
		banksel temp3
		movlw	.0
		movwf	temp3
		movlw	.249
		movwf	temp2
D11		call	rtrn2	
		call	rtrn2
		call	rtrn
		goto	$+1	
		decfsz 	temp3,f
		goto 	D11
		decfsz 	temp2,f
		goto 	D11		
	return


		END